#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "struct_ext.h"
#include "pcg_ext.h"
#include "poi_gen.h"

extern int
POI_GEN(void)
{
	double VOL0, coef;
	int nn, nnp;
	int icN1, icN2, icN3, icN4, icN5, icN6;
	int icel, icou, icouG, i, j, ii, jj, kk, ib;

/*********
 * INIT. *
 *********/
	nn  = ICELTOT;

	NL = 6;
	NU = 6;

	BFORCE = (double *)malloc(sizeof(double)*nn);
	if(BFORCE == NULL) {
		fprintf(stderr, "Error: %s\n", strerror(errno));
		return -1;
	}
	D = (double *)malloc(sizeof(double)*nn);
	if(D == NULL) {
		fprintf(stderr, "Error: %s\n", strerror(errno));
		return -1;
	}
	PHI = (double *)malloc(sizeof(double)*nn);
	if(PHI == NULL) {
		fprintf(stderr, "Error: %s\n", strerror(errno));
		return -1;
	}

	INL = (int *)calloc(nn, sizeof(int));
	if(INL == NULL) {
		fprintf(stderr, "Error: %s\n", strerror(errno));
		return -1;
	}
	INU = (int *)calloc(nn, sizeof(int));
	if(INU == NULL) {
		fprintf(stderr, "Error: %s\n", strerror(errno));
		return -1;
	}

	IAL = (int **)calloc(nn, sizeof(int *));
	if(IAL == NULL) {
		fprintf(stderr, "Error: %s\n", strerror(errno));
		return -1;
	}
	for(i=0; i<nn; i++) {
		IAL[i] = (int *)calloc(NL, sizeof(int));
		if(IAL[i] == NULL) {
			fprintf(stderr, "Error: %s\n", strerror(errno));
			return -1;
		}
	}
	IAU = (int **)calloc(nn, sizeof(int *));
	if(IAU == NULL) {
		fprintf(stderr, "Error: %s\n", strerror(errno));
		return -1;
	}
	for(i=0; i<nn; i++) {
		IAU[i] = (int *)calloc(NU, sizeof(int));
		if(IAU[i] == NULL) {
			fprintf(stderr, "Error: %s\n", strerror(errno));
			return -1;
		}
	}

	AL = (double **)malloc(sizeof(double *)*nn);
	if(AL == NULL) {
		fprintf(stderr, "Error: %s\n", strerror(errno));
		return -1;
	}
	for(i=0; i<nn; i++) {
		AL[i] = (double *)malloc(sizeof(double)*NL);
		if(AL[i] == NULL) {
			fprintf(stderr, "Error: %s\n", strerror(errno));
			return -1;
		}
	}
	AU = (double **)malloc(sizeof(double *)*nn);
	if(AU == NULL) {
		fprintf(stderr, "Error: %s\n", strerror(errno));
		return -1;
	}
	for(i=0; i<nn; i++) {
		AU[i] = (double *)malloc(sizeof(double)*NU);
		if(AU[i] == NULL) {
			fprintf(stderr, "Error: %s\n", strerror(errno));
			return -1;
		}
	}

	for(i=0; i<nn; i++) {
		D[i]   = 0.0;
		PHI[i] = 0.0;
	}
	for(j=0; j<nn; j++) {
		for(i=0; i<NL; i++) {
			AL[j][i] = 0.0;
		}
	}
	for(j=0; j<nn; j++) {
		for(i=0; i<NU; i++) {
			AU[j][i] = 0.0;
		}
	}

/*********************************
 * INTERIOR & NEUMANN boundary's *
 *********************************/
	for(icel=0; icel<ICELTOT; icel++) {
		icN1 = NEIBcell[0][icel];
		icN2 = NEIBcell[1][icel];
		icN3 = NEIBcell[2][icel];
		icN4 = NEIBcell[3][icel];
		icN5 = NEIBcell[4][icel];
		icN6 = NEIBcell[5][icel];

		icouG = 0;


		if(icN5 != 0) {
			icou = INL[icel] + 1;
			IAL[icel][icou-1] = icN5;
			INL[icel]         = icou;
		}

		if(icN3 != 0) {
			icou = INL[icel] + 1;
			IAL[icel][icou-1] = icN3;
			INL[icel]         = icou;
		}

		if(icN1 != 0) {
			icou = INL[icel] + 1;
			IAL[icel][icou-1] = icN1;
			INL[icel]         = icou;
		}

		if(icN2 != 0) {
			icou = INU[icel] + 1;
			IAU[icel][icou-1] = icN2;
			INU[icel]         = icou;
		}

		if(icN4 != 0) {
			icou = INU[icel] + 1;
			IAU[icel][icou-1] = icN4;
			INU[icel]         = icou;
		}

		if(icN6 != 0) {
			icou = INU[icel] + 1;
			IAU[icel][icou-1] = icN6;
			INU[icel]         = icou;
		}
	}

/*************************************
 * INTERIOR & NEUMANN BOUNDARY CELLs *
 *************************************/
	icouG = 0;
	for(icel=0; icel<ICELTOT; icel++) {
		icN1 = NEIBcell[0][icel];
		icN2 = NEIBcell[1][icel];
		icN3 = NEIBcell[2][icel];
		icN4 = NEIBcell[3][icel];
		icN5 = NEIBcell[4][icel];
		icN6 = NEIBcell[5][icel];

		VOL0 = VOLCEL[icel];

		icou = 0;
		if(icN5 != 0) {
			coef = RDZ * ZAREA;
			D[icel] -= coef;
			AL[icel][icou] = coef;
			icou++;
		}

		if(icN3 != 0) {
			coef = RDY * YAREA;
			D[icel] -= coef;
			AL[icel][icou] = coef;
			icou++;
		}

		if(icN1 != 0) {
			coef = RDX * XAREA;
			D[icel] -= coef;
			AL[icel][icou] = coef;
			icou++;
		}

		icou = 0;
		if(icN2 != 0) {
			coef = RDX * XAREA;
			D[icel] -= coef;
			AU[icel][icou] = coef;
			icou++;
		}

		if(icN4 != 0) {
			coef = RDY * YAREA;
			D[icel] -= coef;
			AU[icel][icou] = coef;
			icou++;
		}

		if(icN6 != 0) {
			coef = RDZ * ZAREA;
			D[icel] -= coef;
			AU[icel][icou] = coef;
			icou++;
		}

		ii = XYZ[0][icel];
		jj = XYZ[1][icel];
		kk = XYZ[2][icel];

		BFORCE[icel] = - (double)(ii + jj + kk) * VOL0;
	}

/****************************
 * DIRICHLET BOUNDARY CELLs *
 ****************************/
	for(ib=0; ib<ZmaxCELtot; ib++) {
		icel = ZmaxCEL[ib];
		coef = 2.0 * RDZ * ZAREA;
		D[icel-1] -= coef;
	}

	return 0;
}
