/*
 * solve_ICCG2
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <math.h>

#include "solver_ICCG2.h"

static int
FORM_ILU0(int N, int NL, int NU, int *INL, int **IAL, int *INU, int **IAU,
		double *D, double **AL, double **AU,
		double ***ALlu0, double ***AUlu0, double **Dlu0)
{
	int *IW1, *IW2;
	int *IWsL, *IWsU;
	double RHS_Aij, DkINV, Aik, Akj, D11;
	int ij0, k0;
	int icon, kcon, i, j, k;

/*********
 * INIT. *
 *********/
	*ALlu0 = (double **)malloc(sizeof(double *)*N);
	if(*ALlu0 == NULL) {
		fprintf(stderr, "Error: %s\n", strerror(errno));
		return -1;
	}
	for(i=0; i<N; i++) {
		(*ALlu0)[i] = (double *)malloc(sizeof(double)*NL);
		if((*ALlu0)[i] == NULL) {
			fprintf(stderr, "Error: %s\n", strerror(errno));
			return -1;
		}
	}

	*AUlu0 = (double **)malloc(sizeof(double *)*N);
	if(*AUlu0 == NULL) {
		fprintf(stderr, "Error: %s\n", strerror(errno));
		return -1;
	}
	for(i=0; i<N; i++) {
		(*AUlu0)[i] = (double *)malloc(sizeof(double)*NU);
		if((*AUlu0)[i] == NULL) {
			fprintf(stderr, "Error: %s\n", strerror(errno));
			return -1;
		}
	}

	*Dlu0 = (double *)malloc(sizeof(double)*N);
	if(*Dlu0 == NULL) {
		fprintf(stderr, "Error: %s\n", strerror(errno));
		return -1;
	}

	for(i=0; i<N; i++) {
		(*Dlu0)[i] = D[i];
		for(k=0; k<INL[i]; k++) {
			(*ALlu0)[i][k] = AL[i][k];
		}
		for(k=0; k<INU[i]; k++) {
			(*AUlu0)[i][k] = AU[i][k];
		}
	}

/************************
 * ILU[0] factorization *
 ************************/
	IW1 = (int *)malloc(sizeof(int)*N);
	if(IW1 == NULL) {
		fprintf(stderr, "Error: %s\n", strerror(errno));
		return -1;
	}
	IW2 = (int *)malloc(sizeof(int)*N);
	if(IW2 == NULL) {
		fprintf(stderr, "Error: %s\n", strerror(errno));
		return -1;
	}

	for(i=0; i<N; i++) {
		for(j=0; j<N; j++) {
			IW1[j] = 0;
			IW2[j] = 0;
		}

		for(k0=i; k0<INL[i]; k0++) {
			IW1[IAL[i][k0]-1] = k0;
		}

		for(k0=0; k0<INU[i]; k0++) {
			IW2[IAU[i][k0]-1] = k0;
		}

		for(icon=0; icon<INL[i]; icon++) {
			k = IAL[i][icon] - 1;
			D11 = (*Dlu0)[k];

			DkINV = 1.0 / D11;
			Aik = (*ALlu0)[i][icon];

			for(kcon=0; kcon<INU[k]; kcon++) {
				j = IAU[k][kcon] - 1;

				if(j == i) {
					Akj = (*AUlu0)[k][kcon];
					RHS_Aij = Aik * DkINV * Akj;
					(*Dlu0)[i] -= RHS_Aij;
				}

				if(j < i && IW1[j] != 0) {
					Akj = (*AUlu0)[k][kcon];
					RHS_Aij = Aik * DkINV * Akj;

					ij0 = IW1[j];
					(*ALlu0)[i][ij0] += RHS_Aij;
				}

				if(j > i && IW2[j] != 0) {
					Akj = (*AUlu0)[k][kcon];
					RHS_Aij = Aik * DkINV * Akj;

					ij0 = IW2[j];
					(*AUlu0)[i][ij0] += RHS_Aij;
				}
			}
		}
	}

	for(i=0; i<N; i++) {
		(*Dlu0)[i] = 1.0 / (*Dlu0)[i];
	}

	free(IW1);
	free(IW2);

	return 0;
}


extern int
solve_ICCG2(int N, int NL, int NU, int *INL, int **IAL, int *INU, int **IAU,
		double *D, double *B, double *X, double **AL, double **AU,
		double EPS, int *ITR, int *IER)
{
	double **W, **ALlu0, **AUlu0, *Dlu0;
	double VAL, DNRM2, BNRM2, WVAL, SW, RHO, BETA, RHO1, C1, ALPHA, ERR;
	int i, j, L;
	int R  = 0;
	int Z  = 1;
	int Q  = 1;
	int P  = 2;
	int DD = 3;

/*********
 * INIT. *
 *********/
	W = (double **)malloc(sizeof(double *)*4);
	if(W == NULL) {
		fprintf(stderr, "Error: %s\n", strerror(errno));
		return -1;
	}
	for(i=0; i<4; i++) {
		W[i] = (double *)malloc(sizeof(double)*N);
		if(W[i] == NULL) {
			fprintf(stderr, "Error: %s\n", strerror(errno));
			return -1;
		}
	}
	for(i=0; i<N; i++) {
		X[i]    = 0.0;
		W[0][i] = 0.0;
		W[1][i] = 0.0;
		W[2][i] = 0.0;
		W[3][i] = 0.0;
	}

	FORM_ILU0(N, NL, NU, INL, IAL, INU, IAU, D, AL, AU, &ALlu0, &AUlu0, &Dlu0);

/**************************
 * {r0} = {b} - [A]{xini} *
 **************************/
	for(i=0; i<N; i++) {
		VAL = D[i] * X[i];
		for(j=0; j<INL[i]; j++) {
			VAL += AL[i][j] * X[IAL[i][j]-1];
		}
		for(j=0; j<INU[i]; j++) {
			VAL += AU[i][j] * X[IAU[i][j]-1];
		}
		W[R][i] = B[i] - VAL;
	}

	BNRM2 = 0.0;
	for(i=0; i<N; i++) {
		BNRM2 += pow(B[i], 2.0);
	}

/*************************************************************  ITERATION */
	*ITR = N;

	for(L=0; L<(*ITR); L++) {

/*******************
 * {z} = [Minv]{r} *
 *******************/
		for(i=0; i<N; i++) {
			W[Z][i] = W[R][i];
		}

		for(i=0; i<N; i++) {
			WVAL = W[Z][i];
			for(j=0; j<INL[i]; j++) {
				WVAL = WVAL - ALlu0[i][j] * W[Z][IAL[i][j]-1];
			}
			W[Z][i] = WVAL * Dlu0[i];
		}

		for(i=N-1; i>=0; i--) {
			SW = 0.0;
			for(j=0; j<INU[i]; j++) {
				SW += AUlu0[i][j] * W[Z][IAU[i][j]-1];
			}
			W[Z][i] -= Dlu0[i] * SW;
		}

/****************
 * RHO = {r}{z} *
 ****************/
		RHO = 0.0;
		for(i=0; i<N; i++) {
			RHO += W[R][i] * W[Z][i];
		}

/********************************
 * {p}  = {z}  if     ITER=0    *
 * BETA = RHO / RHO1  otherwise *
 ********************************/
		if(L == 0) {
			for(i=0; i<N; i++) {
				W[P][i] = W[Z][i];
			}
		} else {
			BETA = RHO / RHO1;
			for(i=0; i<N; i++) {
				W[P][i] = W[Z][i] + BETA * W[P][i];
			}
		}

/****************
 * {q} = [A]{p} *
 ****************/
		for(i=0; i<N; i++) {
			VAL = D[i] * W[P][i];
			for(j=0; j<INL[i]; j++) {
				VAL += AL[i][j] * W[P][IAL[i][j]-1];
			}
			for(j=0; j<INU[i]; j++) {
				VAL += AU[i][j] * W[P][IAU[i][j]-1];
			}
			W[Q][i] = VAL;
		}

/************************
 * ALPHA = PHO / {p}{q} *
 ************************/
		C1 = 0.0;
		for(i=0; i<N; i++) {
			C1 += W[P][i] * W[Q][i];
		}

		ALPHA = RHO / C1;

/***************************
 * {x} = {x} + ALPHA * {p} *
 * {r} = {x} - ALPHA * {q} *
 ***************************/
		for(i=0; i<N; i++) {
			X[i]    += ALPHA * W[P][i];
			W[R][i] -= ALPHA * W[Q][i];
		}

		DNRM2 = 0.0;
		for(i=0; i<N; i++) {
			DNRM2 += pow(W[R][i], 2.0);
		}


		ERR = sqrt(DNRM2/BNRM2);
		if(L%1 == 0) {
			fprintf(stderr, "%5d%16.6e\n", L+1, ERR);
		}
		if(ERR < EPS) {
			*IER = 0;
			goto N900;
		} else {
			RHO1 = RHO;
		}
	}
	*IER = 1;

N900:
	fprintf(stderr, "%5d%16.6e\n", L+1, ERR);
	*ITR = L;

	free(W);
	free(Dlu0);
	free(ALlu0);
	free(AUlu0);

	return 0;
}
