#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "struct.h"
#include "pcg.h"
#include "input.h"
#include "pointer_init.h"
#include "boundary_cell.h"
#include "cell_metrics.h"
#include "poi_gen.h"
#include "solver_ICCG_mc.h"
#include "outucd.h"

int
main(int argc, char **argv)
{
	double *WK;
	int NPL, NPU;
	int ISET, ITR, IER;
	int icel, ic0, i;
	double xN, xL, xU;
	double FLOPS;

/*********
 * INIT. *
 *********/
	if(INPUT()) goto error;
	if(POINTER_INIT()) goto error;
	if(BOUNDARY_CELL()) goto error;
	if(CELL_METRICS()) goto error;
	if(POI_GEN()) goto error;

/***************
 * MAIN SOLVER *
 ***************/
	memset(PHI, 0, sizeof(int)*ICELTOT);

	ISET = 0;

	NPL = 0;
	NPU = 0;
	for(i=0; i<ICELTOT; i++) {
		NPL += INL[i];
		NPU += INU[i];
	}

	WK = (double *)malloc(sizeof(double)*ICELTOT);
	if(WK == NULL) {
		fprintf(stderr, "Error: %s\n", strerror(errno));
		goto error;
	}

	if(solve_ICCG_mc(ICELTOT, NL, NU, INL, IAL, INU, IAU,
		D, BFORCE, PHI, AL, AU, NCOLORtot, COLORindex,
		EPSICCG, &ITR, &IER)) goto error;

	xN = (double)ICELTOT;
	xL = (double)NPL;
	xU = (double)NPL;

	FLOPS = xN + 3.0 * xL + 2.0 * (xL + xU + xN) + 2.0 * xN;
	FLOPS += (double)ITR * (4.0 * xL + 16.0 * xN + xN + 4.0 * xU + 2.0) - 2.0 * xN;

	fprintf(stderr, "\nN= %10d\n", ICELTOT);

	for(ic0=0; ic0<ICELTOT; ic0++) {
		icel = NEWtoOLD[ic0];
		WK[icel] = PHI[ic0];
	}

	for(icel=0; icel<ICELTOT; icel++) {
		PHI[icel] = WK[icel];
	}

	if(OUTUCD()) goto error;

	return 0;

error:
	return -1;
}
