#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "struct_ext.h"
#include "pcg_ext.h"
#include "mc.h"
#include "rcm.h"
#include "poi_gen.h"

extern int
POI_GEN(void)
{
	double VOL0;
	double coef;
	int nn, nnp;
	int icN1, icN2, icN3, icN4, icN5, icN6;
	int icel, icou, icouG, icol, ic0, i, j, ii, jj, kk, ib;

/*********
 * INIT. *
 *********/
	nn  = ICELTOT;

	NL = 6;
	NU = 6;

	BFORCE = (double *)malloc(sizeof(double)*nn);
	if(BFORCE == NULL) {
		fprintf(stderr, "Error: %s\n", strerror(errno));
		return -1;
	}
	D = (double *)malloc(sizeof(double)*nn);
	if(D == NULL) {
		fprintf(stderr, "Error: %s\n", strerror(errno));
		return -1;
	}
	PHI = (double *)malloc(sizeof(double)*nn);
	if(PHI == NULL) {
		fprintf(stderr, "Error: %s\n", strerror(errno));
		return -1;
	}

	INL = (int *)calloc(nn, sizeof(int));
	if(INL == NULL) {
		fprintf(stderr, "Error: %s\n", strerror(errno));
		return -1;
	}
	INU = (int *)calloc(nn, sizeof(int));
	if(INU == NULL) {
		fprintf(stderr, "Error: %s\n", strerror(errno));
		return -1;
	}

	IAL = (int **)calloc(NL, sizeof(int *));
	if(IAL == NULL) {
		fprintf(stderr, "Error: %s\n", strerror(errno));
		return -1;
	}
	for(i=0; i<NL; i++) {
		IAL[i] = (int *)calloc(nn, sizeof(int));
		if(IAL[i] == NULL) {
			fprintf(stderr, "Error: %s\n", strerror(errno));
			return -1;
		}
	}
	IAU = (int **)calloc(NU, sizeof(int *));
	if(IAU == NULL) {
		fprintf(stderr, "Error: %s\n", strerror(errno));
		return -1;
	}
	for(i=0; i<NU; i++) {
		IAU[i] = (int *)calloc(nn, sizeof(int));
		if(IAU[i] == NULL) {
			fprintf(stderr, "Error: %s\n", strerror(errno));
			return -1;
		}
	}

	AL = (double **)malloc(sizeof(double *)*NL);
	if(AL == NULL) {
		fprintf(stderr, "Error: %s\n", strerror(errno));
		return -1;
	}
	for(i=0; i<NL; i++) {
		AL[i] = (double *)malloc(sizeof(double)*nn);
		if(AL[i] == NULL) {
			fprintf(stderr, "Error: %s\n", strerror(errno));
			return -1;
		}
	}
	AU = (double **)malloc(sizeof(double *)*NU);
	if(AU == NULL) {
		fprintf(stderr, "Error: %s\n", strerror(errno));
		return -1;
	}
	for(i=0; i<NU; i++) {
		AU[i] = (double *)malloc(sizeof(double)*nn);
		if(AU[i] == NULL) {
			fprintf(stderr, "Error: %s\n", strerror(errno));
			return -1;
		}
	}

	for(i=0; i<nn; i++) {
		D[i] = 0.0;
		PHI[i] = 0.0;
	}
	for(j=0; j<NL; j++) {
		for(i=0; i<nn; i++) {
			AL[j][i] = 0.0;
		}
	}
	for(j=0; j<NU; j++) {
		for(i=0; i<nn; i++) {
			AU[j][i] = 0.0;
		}
	}

/*********************************
 * INTERIOR & NEUMANN boundary's *
 *********************************/
	for(icel=0; icel<ICELTOT; icel++) {
		icN1 = NEIBcell[icel][0];
		icN2 = NEIBcell[icel][1];
		icN3 = NEIBcell[icel][2];
		icN4 = NEIBcell[icel][3];
		icN5 = NEIBcell[icel][4];
		icN6 = NEIBcell[icel][5];

		icouG = 0;


		if(icN5 != 0) {
			icou = INL[icel] + 1;
			IAL[icou-1][icel] = icN5;
			INL        [icel] = icou;
		}

		if(icN3 != 0) {
			icou = INL[icel] + 1;
			IAL[icou-1][icel] = icN3;
			INL        [icel] = icou;
		}

		if(icN1 != 0) {
			icou = INL[icel] + 1;
			IAL[icou-1][icel] = icN1;
			INL        [icel] = icou;
		}

		if(icN2 != 0) {
			icou = INU[icel] + 1;
			IAU[icou-1][icel] = icN2;
			INU        [icel] = icou;
		}

		if(icN4 != 0) {
			icou = INU[icel] + 1;
			IAU[icou-1][icel] = icN4;
			INU        [icel] = icou;
		}

		if(icN6 != 0) {
			icou = INU[icel] + 1;
			IAU[icou-1][icel] = icN6;
			INU        [icel] = icou;
		}
	}

/*****************
 * MULTICOLORING *
 *****************/
	OLDtoNEW = (int *)calloc(ICELTOT, sizeof(int));
	if(OLDtoNEW == NULL) {
		fprintf(stderr, "Error: %s\n", strerror(errno));
		return -1;
	}
	NEWtoOLD = (int *)calloc(ICELTOT, sizeof(int));
	if(NEWtoOLD == NULL) {
		fprintf(stderr, "Error: %s\n", strerror(errno));
		return -1;
	}
	COLORindex = (int *)calloc(ICELTOT+1, sizeof(int));
	if(COLORindex == NULL) {
		fprintf(stderr, "Error: %s\n", strerror(errno));
		return -1;
	}

N111:
	fprintf(stderr, "\n\nYou have%8d elements\n", ICELTOT);
	fprintf(stderr, "How many colors do you need ?\n");
	fprintf(stderr, "  #COLOR must be more than 2 and\n");
	fprintf(stderr, "  #COLOR must not be more than%8d\n", ICELTOT);
	fprintf(stderr, "  if #COLOR=0  : CM  ordering\n");
	fprintf(stderr, "  if #COLOR=-1 : RCM ordering\n");
	fprintf(stderr, "=>\n");
	fscanf(stdin, "%d", &NCOLORtot);
	if(NCOLORtot == 1 || NCOLORtot > ICELTOT) goto N111;

	if(NCOLORtot > 0) {
		MC(ICELTOT, NL, NU, INL, IAL, INU, IAU,
				&NCOLORtot, COLORindex, NEWtoOLD, OLDtoNEW);
	} else if(NCOLORtot == 0) {
		CM(ICELTOT, NL, NU, INL, IAL, INU, IAU,
				&NCOLORtot, COLORindex, NEWtoOLD, OLDtoNEW);
	} else if(NCOLORtot < 0) {
		RCM(ICELTOT, NL, NU, INL, IAL, INU, IAU,
				&NCOLORtot, COLORindex, NEWtoOLD, OLDtoNEW);
	}

	fprintf(stderr, "\n\n### FINAL COLOR NUMBER%8d\n\n", NCOLORtot);

/*************************************
 * INTERIOR & NEUMANN BOUNDARY CELLs *
 *************************************/
	icouG = 0;

	for(icol=0; icol<NCOLORtot; icol++) {
		for(icel=COLORindex[icol]; icel<COLORindex[icol+1]; icel++) {
			ic0  = NEWtoOLD[icel];

			icN1 = NEIBcell[ic0][0];
			icN2 = NEIBcell[ic0][1];
			icN3 = NEIBcell[ic0][2];
			icN4 = NEIBcell[ic0][3];
			icN5 = NEIBcell[ic0][4];
			icN6 = NEIBcell[ic0][5];

			VOL0 = VOLCEL[ic0];

			if(icN5 != 0) {
				icN5 = OLDtoNEW[icN5-1] + 1;
				coef = RDZ * ZAREA;
				D[icel] -= coef;

				if(icN5 < icel+1) {
					for(j=0; j<INL[icel]; j++) {
						if(IAL[j][icel] == icN5) {
							AL[j][icel] = coef;
							break;
						}
					}
				} else {
					for(j=0; j<INU[icel]; j++) {
						if(IAU[j][icel] == icN5) {
							AU[j][icel] = coef;
							break;
						}
					}
				}
			}

			if(icN3 != 0) {
				icN3 = OLDtoNEW[icN3-1] + 1;
				coef = RDY * YAREA;
				D[icel] -= coef;

				if(icN3 < icel+1) {
					for(j=0; j<INL[icel]; j++) {
						if(IAL[j][icel] == icN3) {
							AL[j][icel] = coef;
							break;
						}
					}
				} else {
					for(j=0; j<INU[icel]; j++) {
						if(IAU[j][icel] == icN3) {
							AU[j][icel] = coef;
							break;
						}
					}
				}
			}

			if(icN1 != 0) {
				icN1 = OLDtoNEW[icN1-1] + 1;
				coef = RDX * XAREA;
				D[icel] -= coef;

				if(icN1 < icel+1) {
					for(j=0; j<INL[icel]; j++) {
						if(IAL[j][icel] == icN1) {
							AL[j][icel] = coef;
							break;
						}
					}
				} else {
					for(j=0; j<INU[icel]; j++) {
						if(IAU[j][icel] == icN1) {
							AU[j][icel] = coef;
							break;
						}
					}
				}
			}

			if(icN2 != 0) {
				icN2 = OLDtoNEW[icN2-1] + 1;
				coef = RDX * XAREA;
				D[icel] -= coef;

				if(icN2 < icel+1) {
					for(j=0; j<INL[icel]; j++) {
						if(IAL[j][icel] == icN2) {
							AL[j][icel] = coef;
							break;
						}
					}
				} else {
					for(j=0; j<INU[icel]; j++) {
						if(IAU[j][icel] == icN2) {
							AU[j][icel] = coef;
							break;
						}
					}
				}
			}

			if(icN4 != 0) {
				icN4 = OLDtoNEW[icN4-1] + 1;
				coef = RDY * YAREA;
				D[icel] -= coef;

				if(icN4 < icel+1) {
					for(j=0; j<INL[icel]; j++) {
						if(IAL[j][icel] == icN4) {
							AL[j][icel] = coef;
							break;
						}
					}
				} else {
					for(j=0; j<INU[icel]; j++) {
						if(IAU[j][icel] == icN4) {
							AU[j][icel] = coef;
							break;
						}
					}
				}
			}

			if(icN6 != 0) {
				icN6 = OLDtoNEW[icN6-1] + 1;
				coef = RDZ * ZAREA;
				D[icel] -= coef;

				if(icN6 < icel+1) {
					for(j=0; j<INL[icel]; j++) {
						if(IAL[j][icel] == icN6) {
							AL[j][icel] = coef;
							break;
						}
					}
				} else {
					for(j=0; j<INU[icel]; j++) {
						if(IAU[j][icel] == icN6) {
							AU[j][icel] = coef;
							break;
						}
					}
				}
			}

			ii = XYZ[ic0][0];
			jj = XYZ[ic0][1];
			kk = XYZ[ic0][2];

			BFORCE[icel] = - (double)(ii + jj + kk) * VOL0;
		}
	}

/****************************
 * DIRICHLET BOUNDARY CELLs *
 ****************************/
	for(ib=0; ib<ZmaxCELtot; ib++) {
		ic0 = ZmaxCEL[ib] - 1;
		coef = 2.0 * RDZ * ZAREA;
		icel = OLDtoNEW[ic0];
		D[icel] -= coef;
	}

	return 0;
}
