/*
 * RCM
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "rcm.h"

extern int
RCM(int N, int NL, int NU, int *INL, int **IAL, int *INU, int **IAU,
	int *NCOLORtot, int *COLORindex, int *NEWtoOLD, int *OLDtoNEW)
{
	int **IW;
	int *INLw, *INUw;
	int **IALw, **IAUw;
	int KC, KCT, KCT0, KMIN, KMAX;
	int JC, JN;
	int II;
	int i, j, k, ic;
	int INmin, NODmin;
	int icon, icol, icouG, icou, icou0, in, inC;
    int jL, jU;

/*********
 * INIT. *
 *********/
	IW = (int **)calloc(N, sizeof(int *));
	if(IW == NULL) {
		fprintf(stderr, "Error: %s\n", strerror(errno));
		return -1;
	}
	for(i=0; i<N; i++) {
		IW[i] = (int *)calloc(2, sizeof(int));
	}

	INmin  = N;
	NODmin = -1;

	for(i=0; i<N; i++) {
		icon = 0;
		for(k=0; k<INL[i]; k++) {
			icon++;
		}
		for(k=0; k<INU[i]; k++) {
			icon++;
		}

		if(icon < INmin) {
			INmin  = icon;
			NODmin = i;
		}
	}


	if(NODmin == -1) NODmin = 0;

	IW[NODmin][1] = 1;

	NEWtoOLD[0]      = NODmin;
	OLDtoNEW[NODmin] = 0;

	icol = 0;


/************
 * COLORING *
 ************/
	icouG = 1;
	for(icol=1; icol<N; icol++) {
		icou  = 0;
		icou0 = 0;
		for(i=0; i<N; i++) {
			if(IW[i][1] == icol) {
				for(k=0; k<INL[i]; k++) {
					in = IAL[k][i];
					if(IW[in-1][1] == 0) {
						IW[in-1][1] = -(icol + 1);
						icou++;
						IW[icou-1][0] = in;
					}
				}
				for(k=0; k<INU[i]; k++) {
					in = IAU[k][i];
					if(IW[in-1][1] == 0) {
						IW[in-1][1] = -(icol + 1);
						icou++;
						IW[icou-1][0] = in;
					}
				}
			}
		}

		if(icou == 0) {
			for(i=0; i<N; i++) {
				if(IW[i][1] == 0) {
					icou++;
					IW[i     ][1] = -(icol + 1);
					IW[icou-1][0] = i + 1;
					break;
				}
			}
		}

		for(ic=0; ic<icou; ic++) {
			inC = IW[ic][0];
			if(IW[inC-1][1] != 0) {
				for(k=0; k<INL[inC-1]; k++) {
					in = IAL[k][inC-1];
					if(IW[in-1][1] <= 0) {
						IW[in-1][1] = 0;
					}
				}
				for(k=0; k<INU[inC-1]; k++) {
					in = IAU[k][inC-1];
					if(IW[in-1][1] <= 0) {
						IW[in-1][1] = 0;
					}
				}
			}
		}

		for(ic=0; ic<icou; ic++) {
			inC = IW[ic][0];
			if(IW[inC-1][1] != 0) {
				icouG++;
				IW[inC-1][1] = icol + 1;
			}
		}

		if(icouG == N) break;
	}

/******************/
/* FINAL COLORING */
/******************/
	*NCOLORtot = icol + 1;
	icouG = 0;
	for(ic=(*NCOLORtot); ic>0; ic--) {
		icou = 0;
		for(i=N-1; i>=0; i--) {
			if(IW[i][1] == ic) {
				NEWtoOLD[icouG] = i;
				OLDtoNEW[i    ] = icouG;
				icou++;
				icouG++;
			}
		}
		COLORindex[*NCOLORtot-ic+1] = icou;
	}

	COLORindex[0] = 0;
	for(ic=1; ic<=(*NCOLORtot); ic++) {
		COLORindex[ic] = COLORindex[ic-1] + COLORindex[ic];
	}

/*******************/
/* MATRIX transfer */
/*******************/
	INLw = (int *)calloc(N, sizeof(int));
	if(INLw == NULL) {
		fprintf(stderr, "Error: %s\n", strerror(errno));
		return -1;
	}
	INUw = (int *)calloc(N, sizeof(int));
	if(INUw == NULL) {
		fprintf(stderr, "Error: %s\n", strerror(errno));
		return -1;
	}
	IALw = (int **)calloc(NL, sizeof(int *));
	if(IALw == NULL) {
		fprintf(stderr, "Error: %s\n", strerror(errno));
		return -1;
	}
	for(i=0; i<NL; i++) {
		IALw[i] = (int *)calloc(N, sizeof(int));
		if(IALw[i] == NULL) {
			fprintf(stderr, "Error: %s\n", strerror(errno));
			return -1;
		}
	}
	IAUw = (int **)calloc(NU, sizeof(int *));
	if(IAUw == NULL) {
		fprintf(stderr, "Error: %s\n", strerror(errno));
		return -1;
	}
	for(i=0; i<NU; i++) {
		IAUw[i] = (int *)calloc(N, sizeof(int *));
		if(IAUw[i] == NULL) {
			fprintf(stderr, "Error: %s\n", strerror(errno));
			return -1;
		}
	}

	for(j=0; j<NL; j++) {
		for(i=0; i<N; i++) {
			IW[i][0] = IAL[j][NEWtoOLD[i]];
		}
		for(i=0; i<N; i++) {
			IAL[j][i] = IW[i][0];
		}
	}

	for(j=0; j<NU; j++) {
		for(i=0; i<N; i++) {
			IW[i][0] = IAU[j][NEWtoOLD[i]];
		}
		for(i=0; i<N; i++) {
			IAU[j][i] = IW[i][0];
		}
	}

	for(i=0; i<N; i++) {
		IW[i][0] = INL[NEWtoOLD[i]];
	}

	for(i=0; i<N; i++) {
		INLw[i] = IW[i][0];
	}

	for(i=0; i<N; i++) {
		IW[i][0] = INU[NEWtoOLD[i]];
	}

	for(i=0; i<N; i++) {
		INUw[i] = IW[i][0];
	}

	for(j=0; j<NL; j++) {
		for(i=0; i<N; i++) {
			if(IAL[j][i] == 0) {
				IALw[j][i] = 0;
			} else {
				IALw[j][i] = OLDtoNEW[IAL[j][i]-1] + 1;
			}
		}
	}

	for(j=0; j<NU; j++) {
		for(i=0; i<N; i++) {
			if(IAU[j][i] == 0) {
				IAUw[j][i] = 0;
			} else {
				IAUw[j][i] = OLDtoNEW[IAU[j][i]-1] + 1;
			}
		}
	}

	memset(INL, 0, sizeof(int)*N);
	memset(INU, 0, sizeof(int)*N);
	for(i=0; i<NL; i++) {
		memset(IAL[i], 0, sizeof(int)*N);
	}
	for(i=0; i<NU; i++) {
		memset(IAU[i], 0, sizeof(int)*N);
	}

	for(i=0; i<N; i++) {
		jL = 0;
		jU = 0;

		for(j=0; j<INLw[i]; j++) {
			if(IALw[j][i] > i+1) {
				IAU[jU][i] = IALw[j][i];
				jU++;
			} else {
				IAL[jL][i] = IALw[j][i];
				jL++;
			}
		}

		for(j=0; j<INUw[i]; j++) {
			if(IAUw[j][i] > i+1) {
				IAU[jU][i] = IAUw[j][i];
				jU++;
			} else {
				IAL[jL][i] = IAUw[j][i];
				jL++;
			}
		}

		INL[i] = jL;
		INU[i] = jU;
	}

	free(IW);
	free(INLw);
	free(INUw);
	free(IALw);
	free(IAUw);

	return 0;
}
