/*
 * solver_ICCG
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <math.h>

#include "solver_ICCG_mc.h"

extern int
solve_ICCG_mc(int N, int NL, int NU, int *INL, int **IAL, int *INU, int **IAU,
		double *D, double *B, double *X, double **AL, double **AU,
		int NCOLORtot, int *COLORindex, double EPS, int *ITR, int *IER)
{
	double **W;
	double VAL, BNRM2, WVAL, SW, RHO, BETA, RHO1, C1, DNRM2, ALPHA, ERR;
	int i, j, L, ic;
	int R = 0;
	int Z = 1;
	int Q = 1;
	int P = 2;
	int DD = 3;

/*********
 * INIT. *
 *********/
	W = (double **)malloc(sizeof(double *)*N);
	if(W == NULL) {
		fprintf(stderr, "Error: %s\n", strerror(errno));
		return -1;
	}
	for(i=0; i<N; i++) {
		W[i] = (double *)malloc(sizeof(double)*4);
		if(W[i] == NULL) {
			fprintf(stderr, "Error: %s\n", strerror(errno));
			return -1;
		}
	}
	for(i=0; i<N; i++) {
		X[i] = 0.0;
		W[i][0] = 0.0;
		W[i][1] = 0.0;
		W[i][2] = 0.0;
		W[i][3] = 0.0;
	}
	for(ic=0; ic<NCOLORtot; ic++) {
		for(i=COLORindex[ic]; i<COLORindex[ic+1]; i++) {
			VAL = D[i];
			for(j=0; j<INL[i]; j++) {
				VAL = VAL - pow(AL[j][i], 2.0) * W[IAL[j][i]-1][DD];
			}
			W[i][DD] = 1.0 / VAL;
		}
	}

/**************************
 * {r0} = {b} - {A}{xini} *
 **************************/
		for(i=0; i<N; i++) {
			VAL = D[i] * X[i];
			for(j=0; j<INL[i]; j++) {
				VAL += AL[j][i] * X[IAL[j][i]-1];
			}
			for(j=0; j<INU[i]; j++) {
				VAL += AU[j][i] * X[IAU[j][i]-1];
			}
			W[i][R] = B[i] - VAL;
		}

	BNRM2 = 0.0;
	for(i=0; i<N; i++) {
		BNRM2 += pow(B[i], 2.0);
	}

/************************************************************** ITERATION */
	*ITR = N;

	for(L=0; L<(*ITR); L++) {

/*******************
 * {z} = [Minv]{r} *
 *******************/
		for(i=0; i<N; i++) {
			W[i][Z] = W[i][R];
		}

		for(ic=0; ic<NCOLORtot; ic++) {
			for(i=COLORindex[ic]; i<COLORindex[ic+1]; i++) {
				WVAL = W[i][Z];
				for(j=0; j<INL[i]; j++) {
					WVAL -= AL[j][i] * W[IAL[j][i]-1][Z];
				}
				W[i][Z] = WVAL * W[i][DD];
			}
		}

		for(ic=NCOLORtot-1; ic>=0; ic--) {
			for(i=COLORindex[ic]; i<COLORindex[ic+1]; i++) {
				SW = 0.0;
				for(j=0; j<INU[i]; j++) {
					SW += AU[j][i] * W[IAU[j][i]-1][Z];
				}
				W[i][Z] -= W[i][DD] * SW;
			}
		}

/****************
 * RHO = {r}{z} *
 ****************/
		RHO = 0.0;
		for(i=0; i<N; i++) {
			RHO += W[i][R] * W[i][Z];
		}

/********************************
 * {p}  = {z} if      ITER=0    *
 * BETA = RHO / RHO1  otherwise *
 ********************************/
		if(L == 0) {
			for(i=0; i<N; i++) {
				W[i][P] = W[i][Z];
			}
		} else {
			BETA = RHO / RHO1;
			for(i=0; i<N; i++) {
				W[i][P] = W[i][Z] + BETA * W[i][P];
			}
		}

/****************
 * {q} = [A]{p} *
 ****************/
			for(i=0; i<N; i++) {
				VAL = D[i] * W[i][P];
				for(j=0; j<INL[i]; j++) {
					VAL += AL[j][i] * W[IAL[j][i]-1][P];
				}
				for(j=0; j<INU[i]; j++) {
					VAL += AU[j][i] * W[IAU[j][i]-1][P];
				}
				W[i][Q] = VAL;
			}

/************************
 * ALPHA = RHO / {p}{q} *
 ************************/
		C1 = 0.0;
		for(i=0; i<N; i++) {
			C1 += W[i][P] * W[i][Q];
		}

		ALPHA = RHO / C1;

/***************************
 * {x} = {x} + ALPHA * {p} *
 * {r} = {r} - ALPHA * {q} *
 ***************************/
		for(i=0; i<N; i++) {
			X[i]    += ALPHA * W[i][P];
			W[i][R] -= ALPHA * W[i][Q];
		}

		DNRM2 = 0.0;
		for(i=0; i<N; i++) {
			DNRM2 += pow(W[i][R], 2.0);
		}

		ERR = sqrt(DNRM2/BNRM2);
		if(L%1 == 0) {
			fprintf(stderr, "%5d%16.6e\n", L+1, ERR);
		}
		if(ERR < EPS) {
			*IER = 0;
			goto N900;
		} else {
			RHO1 = RHO;
		}
	}
	*IER = 1;

N900:
	fprintf(stderr, "%5d%16.6e\n", L+1, ERR);

	*ITR = L;

	free(W);

	return 0;
}
