#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "struct.h"
#include "pcg.h"
#include "input.h"
#include "pointer_init.h"
#include "boundary_cell.h"
#include "cell_metrics.h"
#include "poi_gen.h"
#include "solver_ICCG_mc.h"
#include "outucd.h"

int
main(int argc, char **argv)
{
	double *WK;
	int NPL, NPU;
	int ISET, ITR, IER;
	int icel, ic0, i;
	double xN, xL, xU;
	double Stime, Etime;

/*************
 * init. MPI *
 *************/
	MPI_Init(&argc, &argv);

/*********
 * INIT. *
 *********/
	if(INPUT()) goto error;
	if(POINTER_INIT()) goto error;
	if(BOUNDARY_CELL()) goto error;
	if(CELL_METRICS()) goto error;
	if(POI_GEN()) goto error;

/***************
 * MAIN SOLVER *
 ***************/
	memset(PHI, 0, sizeof(double)*ICELTOT);

	ISET = 0;

	WK = (double *)malloc(sizeof(double)*ICELTOT);
	if(WK == NULL) {
		fprintf(stderr, "Error: %s\n", strerror(errno));
		goto error;
	}

	Stime = MPI_Wtime();
	if(solve_ICCG_mc(ICELTOT, NL, NU, INL, IAL, INU, IAU,
		D, BFORCE, PHI, AL, AU, NCOLORtot, PEsmpTOT,
		SMPindex, SMPindexG, EPSICCG, &ITR, &IER)) goto error;
	Etime = MPI_Wtime();

	fprintf(stderr, "\nN= %10d\n", ICELTOT);
	fprintf(stderr, "%16.6e sec. (solver)\n", Etime - Stime);

	for(ic0=0; ic0<ICELTOT; ic0++) {
		icel = NEWtoOLD[ic0];
		WK[icel] = PHI[ic0];
	}

	for(icel=0; icel<ICELTOT; icel++) {
		PHI[icel] = WK[icel];
	}

	if(OUTUCD()) goto error;

	MPI_Finalize();

	return 0;

error:
	return -1;
}
