/*
 * solver_ICCG
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <math.h>

#include "solver_ICCG_mc.h"

extern int
solve_ICCG_mc(int N, int NL, int NU, int *INL, int **IAL, int *INU, int **IAU,
		double *D, double *B, double *X, double **AL, double **AU,
		int NCOLORtot, int PEsmpTOT, int *SMPindex, int *SMPindexG,
		double EPS, int *ITR, int *IER)
{
	double **W;
	double VAL, BNRM2, WVAL, SW, RHO, BETA, RHO1, C1, DNRM2, ALPHA, ERR;
	int i, j, ic, ip, L, ip1;
	int R = 0;
	int Z = 1;
	int Q = 1;
	int P = 2;
	int DD = 3;

/*********
 * INIT. *
 *********/
	W = (double **)malloc(sizeof(double *)*4);
	if(W == NULL) {
		fprintf(stderr, "Error: %s\n", strerror(errno));
		return -1;
	}
	for(i=0; i<4; i++) {
		W[i] = (double *)malloc(sizeof(double)*N);
		if(W[i] == NULL) {
			fprintf(stderr, "Error: %s\n", strerror(errno));
			return -1;
		}
	}
#pragma omp parallel for private (ip, i)
	for(ip=0; ip<PEsmpTOT; ip++) {
		for(i=SMPindexG[ip]; i<SMPindexG[ip+1]; i++) {
			X[i] = 0.0;
/*			W[0][i] = 0.0; */
			W[1][i] = 0.0;
			W[2][i] = 0.0;
			W[3][i] = 0.0;
		}
	}

	for(ic=0; ic<NCOLORtot; ic++) {
#pragma omp parallel for private (ip, ip1, i, VAL, j)
		for(ip=0; ip<PEsmpTOT; ip++) {
			ip1 = ic * PEsmpTOT + ip;
			for(i=SMPindex[ip1]; i<SMPindex[ip1+1]; i++) {
				VAL = D[i];
				for(j=0; j<INL[i]; j++) {
					VAL = VAL - pow(AL[i][j], 2.0) * W[DD][IAL[i][j] - 1];
				}
				W[DD][i] = 1.0 / VAL;
			}
		}
	}

/**************************
 * {r0} = {b} - {A}{xini} *
 **************************/
#pragma omp parallel for private (ip, i, VAL, j)
	for(ip=0; ip<PEsmpTOT; ip++) {
		for(i=SMPindexG[ip]; i<SMPindexG[ip+1]; i++) {
			VAL = D[i] * X[i];
			for(j=0; j<INL[i]; j++) {
				VAL += AL[i][j] * X[IAL[i][j]-1];
			}
			for(j=0; j<INU[i]; j++) {
				VAL += AU[i][j] * X[IAU[i][j]-1];
			}
			W[R][i] = B[i] - VAL;
		}
	}

	BNRM2 = 0.0;
#pragma omp parallel for private (ip, i) reduction (+:BNRM2)
	for(ip=0; ip<PEsmpTOT; ip++) {
		for(i=SMPindexG[ip]; i<SMPindexG[ip+1]; i++) {
			BNRM2 += pow(B[i], 2.0);
		}
	}

/************************************************************** ITERATION */
	*ITR = N;

	for(L=0; L<(*ITR); L++) {

/*******************
 * {z} = [Minv]{r} *
 *******************/
#pragma omp parallel for private(ip, i)
		for(ip=0; ip<PEsmpTOT; ip++) {
			for(i=SMPindexG[ip]; i<SMPindexG[ip+1]; i++) {
				W[Z][i] = W[R][i];
			}
		}

		for(ic=0; ic<NCOLORtot; ic++) {
#pragma omp parallel for private (ip, ip1, i, WVAL, j)
			for(ip=0; ip<PEsmpTOT; ip++) {
				ip1 = ic * PEsmpTOT + ip;
				for(i=SMPindex[ip1]; i<SMPindex[ip1+1]; i++) {
					WVAL = W[Z][i];
					for(j=0; j<INL[i]; j++) {
						WVAL -= AL[i][j] * W[Z][IAL[i][j]-1];
					}
					W[Z][i] = WVAL * W[DD][i];
				}
			}
		}

		for(ic=NCOLORtot-1; ic>=0; ic--) {
#pragma omp parallel for private (ip, ip1, i, SW, j)
			for(ip=0; ip<PEsmpTOT; ip++) {
				ip1 = ic * PEsmpTOT + ip;
				for(i=SMPindex[ip1]; i<SMPindex[ip1+1]; i++) {
					SW = 0.0;
					for(j=0; j<INU[i]; j++) {
						SW += AU[i][j] * W[Z][IAU[i][j]-1];
					}
					W[Z][i] -= W[DD][i] * SW;
				}
			}
		}

/****************
 * RHO = {r}{z} *
 ****************/
		RHO = 0.0;
#pragma omp parallel for private (ip, i) reduction (+:RHO)
		for(ip=0; ip<PEsmpTOT; ip++) {
			for(i=SMPindexG[ip]; i<SMPindexG[ip+1]; i++) {
				RHO += W[R][i] * W[Z][i];
			}
		}

/********************************
 * {p}  = {z} if      ITER=0    *
 * BETA = RHO / RHO1  otherwise *
 ********************************/
		if(L == 0) {
#pragma omp parallel for private (ip, i)
			for(ip=0; ip<PEsmpTOT; ip++) {
				for(i=SMPindexG[ip]; i<SMPindexG[ip+1]; i++) {
					W[P][i] = W[Z][i];
				}
			}
		} else {
			BETA = RHO / RHO1;
#pragma omp parallel for private (ip, i)
			for(ip=0; ip<PEsmpTOT; ip++) {
				for(i=SMPindexG[ip]; i<SMPindexG[ip+1]; i++) {
					W[P][i] = W[Z][i] + BETA * W[P][i];
				}
			}
		}

/****************
 * {q} = [A]{p} *
 ****************/
#pragma omp parallel for private (ip, i, VAL, j)
		for(ip=0; ip<PEsmpTOT; ip++) {
			for(i=SMPindexG[ip]; i<SMPindexG[ip+1]; i++) {
				VAL = D[i] * W[P][i];
				for(j=0; j<INL[i]; j++) {
					VAL += AL[i][j] * W[P][IAL[i][j]-1];
				}
				for(j=0; j<INU[i]; j++) {
					VAL += AU[i][j] * W[P][IAU[i][j]-1];
				}
				W[Q][i] = VAL;
			}
		}

/************************
 * ALPHA = RHO / {p}{q} *
 ************************/
		C1 = 0.0;
#pragma omp parallel for private (ip, i) reduction (+:C1)
		for(ip=0; ip<PEsmpTOT; ip++) {
			for(i=SMPindexG[ip]; i<SMPindexG[ip+1]; i++) {
				C1 += W[P][i] * W[Q][i];
			}
		}

		ALPHA = RHO / C1;

/***************************
 * {x} = {x} + ALPHA * {p} *
 * {r} = {r} - ALPHA * {q} *
 ***************************/
#pragma omp parallel for private (ip, i)
		for(ip=0; ip<PEsmpTOT; ip++) {
			for(i=SMPindexG[ip]; i<SMPindexG[ip+1]; i++) {
				X[i]    += ALPHA * W[P][i];
				W[R][i] -= ALPHA * W[Q][i];
			}
		}

		DNRM2 = 0.0;
#pragma omp parallel for private (ip, i) reduction (+:DNRM2)
		for(ip=0; ip<PEsmpTOT; ip++) {
			for(i=SMPindexG[ip]; i<SMPindexG[ip+1]; i++) {
				DNRM2 += pow(W[R][i], 2.0);
			}
		}

		ERR = sqrt(DNRM2/BNRM2);
		if(ERR < EPS) {
			*IER = 0;
			goto N900;
		} else {
			RHO1 = RHO;
		}
	}
	*IER = 1;

N900:
	fprintf(stderr, "%5d%16.6e\n", L+1, ERR);
	*ITR = L;

	free(W);

	return 0;
}
