!C
!C***
!C*** POI_GEN
!C***
!C
!C    generate COEF. MATRIX for POISSON equations
!C    
      subroutine POI_GEN

      use STRUCT
      use PCG

      implicit REAL*8 (A-H,O-Z)

!C
!C-- INIT.
      nn = ICELTOT

      NL= 6
      NU= 6

      allocate (BFORCE(nn), D(nn), PHI(nn))
      allocate (INL(nn), INU(nn), IAL(NL,nn), IAU(NU,nn))
      allocate (                   AL(NL,nn),  AU(NU,nn))

      if (NFLAG.eq.0) then
          BFORCE= 0.d0
          PHI   = 0.d0
          D     = 0.d0
          INL= 0
          INU= 0
          IAL= 0
          IAU= 0
           AL= 0.d0
           AU= 0.d0
      else
!$omp parallel do private (ip,icel,k) 
      do ip= 1, PEsmpTOT
        do icel= ELMindex(ip,0)+1, ELMindex(ip,LEVELtot)
          BFORCE(icel)= 0.d0
          PHI   (icel)= 0.d0
          D(icel)     = 0.d0
          INL(icel)= 0
          INU(icel)= 0
          do k= 1, NL
            IAL(k,icel)= 0            
             AL(k,icel)= 0.d0
          enddo
          do k= 1, NU
            IAU(k,icel)= 0            
             AU(k,icel)= 0.d0
          enddo
        enddo
      enddo
!$omp end parallel do
      endif

!C
!C-- INTERIOR & NEUMANN boundary's
      do icel= 1, ICELTOT
        id0 = LEVEL(icel)
        icN1= NEIBcell(icel,1)
        icN2= NEIBcell(icel,2)
        icN3= NEIBcell(icel,3)
        icN4= NEIBcell(icel,4)
        icN5= NEIBcell(icel,5)
        icN6= NEIBcell(icel,6)

        if (icN5.ne.0) then
          id5= LEVEL(icN5)
          call FIND_CELL (icN5, id5, icel, id0, 5)
        endif

        if (icN3.ne.0) then
          id3= LEVEL(icN3)
          call FIND_CELL (icN3, id3, icel, id0, 3)
        endif

        if (icN1.ne.0) then
          id1= LEVEL(icN1)
          call FIND_CELL (icN1, id1, icel, id0, 1)
        endif

        if (icN2.ne.0) then
          id2= LEVEL(icN2)
          call FIND_CELL (icN2, id2, icel, id0, 2)
        endif

        if (icN4.ne.0) then
          id4= LEVEL(icN4)
          call FIND_CELL (icN4, id4, icel, id0, 4)
        endif

        if (icN6.ne.0) then
          id6= LEVEL(icN6)
          call FIND_CELL (icN6, id6, icel, id0, 6)
        endif

      enddo
!C===

!C
!C +-----------------------------------+
!C | INTERIOR & NEUMANN BOUNDARY CELLs |
!C +-----------------------------------+
!C===
      S1t= MPI_WTIME()
!$omp parallel do private (ip,icel,icN1,icN2,icN3,icN4, icN5,icN6)      &
!$omp&            private (coef,j,ii,jj,kk)
      do ip  = 1, PEsmpTOT
      do icel= ELMindex(ip,0)+1, ELMindex(ip,LEVELtot)
        icN1= NEIBcell(icel,1)
        icN2= NEIBcell(icel,2)
        icN3= NEIBcell(icel,3)
        icN4= NEIBcell(icel,4)
        icN5= NEIBcell(icel,5)
        icN6= NEIBcell(icel,6)

        if (icN5.ne.0) then
          coef= RDZ * ZAREA
          D(icel)= D(icel) - coef

          if (icN5.lt.0) then
            do j= 1, INL(icel)
              if (IAL(j,icel).eq.-icN5) then
                AL(j,icel)= coef
                exit
              endif
            enddo
           else
            do j= 1, INU(icel)
              if (IAU(j,icel).eq.icN5) then
                AU(j,icel)= coef
                exit
              endif
            enddo
          endif

        endif

        if (icN3.ne.0) then
          coef= RDY * YAREA
          D(icel)= D(icel) - coef

          if (icN3.lt.0) then
            do j= 1, INL(icel)
              if (IAL(j,icel).eq.-icN3) then
                AL(j,icel)= coef
                exit
              endif
            enddo
           else
            do j= 1, INU(icel)
              if (IAU(j,icel).eq.icN3) then
                AU(j,icel)= coef
                exit
              endif
            enddo
          endif
        endif

        if (icN1.ne.0) then
          coef= RDX * XAREA
          D(icel)= D(icel) - coef

          if (icN1.lt.0) then
            do j= 1, INL(icel)
              if (IAL(j,icel).eq.-icN1) then
                AL(j,icel)= coef
                exit
              endif
            enddo
           else
            do j= 1, INU(icel)
              if (IAU(j,icel).eq.icN1) then
                AU(j,icel)= coef
                exit
              endif
            enddo
          endif
        endif

        if (icN2.ne.0) then
          coef= RDX * XAREA
          D(icel)= D(icel) - coef

          if (icN2.lt.0) then
            do j= 1, INL(icel)
              if (IAL(j,icel).eq.-icN2) then
                AL(j,icel)= coef
                exit
              endif
            enddo
           else
            do j= 1, INU(icel)
              if (IAU(j,icel).eq.icN2) then
                AU(j,icel)= coef
                exit
              endif
            enddo
          endif
        endif

        if (icN4.ne.0) then
          coef= RDY * YAREA
          D(icel)= D(icel) - coef

          if (icN4.lt.0) then
            do j= 1, INL(icel)
              if (IAL(j,icel).eq.-icN4) then
                AL(j,icel)= coef
                exit
              endif
            enddo
           else
            do j= 1, INU(icel)
              if (IAU(j,icel).eq.icN4) then
                AU(j,icel)= coef
                exit
              endif
            enddo
          endif
        endif

        if (icN6.ne.0) then
          coef= RDZ * ZAREA
          D(icel)= D(icel) - coef

          if (icN6.lt.0) then
            do j= 1, INL(icel)
              if (IAL(j,icel).eq.-icN6) then
                AL(j,icel)= coef
                exit
              endif
            enddo
           else
            do j= 1, INU(icel)
              if (IAU(j,icel).eq.icN6) then
                AU(j,icel)= coef
                exit
              endif
            enddo
          endif
        endif

        ii= XYZ(icel,1)
        jj= XYZ(icel,2)
        kk= XYZ(icel,3)

        BFORCE(icel)= -dfloat(ii+jj+kk) * VOLCEL(icel)

      enddo
      enddo
!$omp end parallel do

      E1t= MPI_WTIME()
      write (*,'(1pe16.6, a)') E1t-S1t, ' sec. (assemble)'
!C===

!C
!C +--------------------------+
!C | DIRICHLET BOUNDARY CELLs |
!C +--------------------------+
!C   TOP SURFACE
!C===
      do ib= 1, ZmaxCELtot
        icel= ZmaxCEL(ib)
        coef= 2.d0 * RDZ * ZAREA
        D(icel)= D(icel) - coef
      enddo
!C===
      contains

!C
!C***
!C*** FIND_CELL
!C***
!C
      subroutine FIND_CELL (icNK, idK, icel, id0, kk)

      if (idK.lt.id0) then
        icou= INL(icel) + 1
        IAL(icou,icel)= icNK
        INL(     icel)= icou
        NEIBcell(icel,kk)= -NEIBcell(icel,kk)
      endif

      if (idK.gt.id0) then
        icou= INU(icel) + 1
        IAU(icou,icel)= icNK
        INU(     icel)= icou
      endif
      if (idK.eq.id0) then
        if (icNK.lt.icel) then
          icou= INL(icel) + 1
          IAL(icou,icel)= icNK
          INL(     icel)= icou
          NEIBcell(icel,kk)= -NEIBcell(icel,kk)
         else
          icou= INU(icel) + 1
          IAU(icou,icel)= icNK
          INU(     icel)= icou
        endif
      endif

      end subroutine FIND_CELL

      end subroutine POI_GEN
