!C*** 
!C*** module solver_ICCG_mc
!C***
!
      module solver_ICCG_mc
      contains
!C
!C*** solve_ICCG
!C
      subroutine solve_ICCG_mc                                          &
     &         ( N, NL, NU, INL, IAL, INU, IAU, D, B, X,                &
     &           AL, AU, LEVELtot, PEsmpTOT, ELMindex,                  &
     &           EPS, ITR, IER)

      implicit REAL*8 (A-H,O-Z)

      integer :: N, NL, NU, LEVELtot, PEsmpTOT

      real(kind=8), dimension(N)   :: D
      real(kind=8), dimension(N)   :: B
      real(kind=8), dimension(N)   :: X
      real(kind=8), dimension(NL,N) :: AL
      real(kind=8), dimension(NU,N) :: AU

      integer, dimension(N)   :: INL, INU
      integer, dimension(NL,N):: IAL
      integer, dimension(NU,N):: IAU
      integer, dimension(PEsmpTOT,0:LEVELtot):: ELMindex

      real(kind=8), dimension(:,:), allocatable :: W

      integer, parameter ::  R= 1
      integer, parameter ::  Z= 2
      integer, parameter ::  Q= 2
      integer, parameter ::  P= 3
      integer, parameter :: DD= 4

!C
!C +------+
!C | INIT |
!C +------+
!C===
      allocate (W(N,4))

!$omp parallel do private(ip,i)
      do ip= 1, PEsmpTOT
      do i = ELMindex(ip,0)+1, ELMindex(ip,LEVELtot)
        X(i)  = 0.d0
        W(i,2)= 0.0D0
        W(i,3)= 0.0D0
        W(i,4)= 0.0D0
      enddo
      enddo     
!$omp end parallel do

      do ic= 1, LEVELtot
!$omp parallel do private(ip,i,VAL,j)
        do ip= 1, PEsmpTOT
        do i= ELMindex(ip,ic-1)+1, ELMindex(ip,ic)
          VAL= D(i)
          do j= 1, INL(i)
            VAL= VAL - (AL(j,i)**2) * W(IAL(j,i),DD)
          enddo
          W(i,DD)= 1.d0/VAL
        enddo
        enddo
!$omp end parallel do
      enddo
!C===

!C
!C +-----------------------+
!C | {r0}= {b} - [A]{xini} |
!C +-----------------------+
!C===

!$omp parallel do private(ip,i,VAL,j)
      do ip= 1, PEsmpTOT
        do i = ELMindex(ip,0)+1, ELMindex(ip,LEVELtot)
          VAL= D(i)*X(i)
          do j= 1, INL(i)
            VAL= VAL + AL(j,i)*X(IAL(j,i))
          enddo 
          do j= 1, INU(i)
            VAL= VAL + AU(j,i)*X(IAU(j,i))
          enddo 
          W(i,R)= B(i) - VAL
        enddo
        enddo
!$omp end parallel do

      BNRM2= 0.0D0
!$omp parallel do private(ip,i) reduction(+:BNRM2)
      do ip= 1, PEsmpTOT
      do i = ELMindex(ip,0)+1, ELMindex(ip,LEVELtot)
        BNRM2 = BNRM2 + B(i)  **2
      enddo
      enddo
!$omp end parallel do
!C===

!C
!C***************************************************************  ITERATION
      ITR= N

      do L= 1, ITR
!C
!C +----------------+
!C | {z}= [Minv]{r} |
!C +----------------+
!C===      

!$omp parallel do private(ip,i)
      do ip= 1, PEsmpTOT
      do i = ELMindex(ip,0)+1, ELMindex(ip,LEVELtot)
        W(i,Z)= W(i,R)
      enddo
      enddo
!$omp end parallel do

      do ic= 1, LEVELtot
!$omp parallel do private(ip,i,WVAL,j)
        do ip= 1, PEsmpTOT
        do i= ELMindex(ip,ic-1)+1, ELMindex(ip,ic)
          WVAL= W(i,Z)
          do j= 1, INL(i)
            WVAL= WVAL -  AL(j,i) * W(IAL(j,i),Z)
          enddo
          W(i,Z)= WVAL * W(i,DD)
        enddo
        enddo
!$omp end parallel do
      enddo

      do ic= LEVELtot, 1, -1
!$omp parallel do private(ip,i,SW,j)
        do ip= 1, PEsmpTOT
        do i= ELMindex(ip,ic), ELMindex(ip,ic-1)+1, -1 
          SW  = 0.0d0
          do j= 1, INU(i)
            SW= SW + AU(j,i) * W(IAU(j,i),Z)
          enddo
          W(i,Z)= W(i,Z) - W(i,DD) * SW
        enddo
        enddo
!$omp end parallel do
      enddo
!C===

!C
!C +-------------+
!C | RHO= {r}{z} |
!C +-------------+
!C===
      RHO= 0.d0
!$omp parallel do private(ip,i) reduction(+:RHO)
      do ip= 1, PEsmpTOT
      do i = ELMindex(ip,0)+1, ELMindex(ip,LEVELtot)
        RHO= RHO + W(i,R)*W(i,Z)   
      enddo          
      enddo          
!$omp end parallel do
!C===

!C
!C +-----------------------------+
!C | {p} = {z} if      ITER=1    |
!C | BETA= RHO / RHO1  otherwise |
!C +-----------------------------+
!C===
      if ( L.eq.1 ) then
!$omp parallel do private(ip,i)
        do ip= 1, PEsmpTOT
        do i = ELMindex(ip,0)+1, ELMindex(ip,LEVELtot)
          W(i,P)= W(i,Z)
        enddo
        enddo
!$omp end parallel do
       else
        BETA= RHO / RHO1
!$omp parallel do private(ip,i)
        do ip= 1, PEsmpTOT
        do i = ELMindex(ip,0)+1, ELMindex(ip,LEVELtot)
          W(i,P)= W(i,Z) + BETA*W(i,P)
        enddo
        enddo
!$omp end parallel do
      endif
!C===

!C
!C +-------------+
!C | {q}= [A]{p} |
!C +-------------+
!C===        
!$omp parallel do private(ip,i,VAL,j)
      do ip= 1, PEsmpTOT
        do i = ELMindex(ip,0)+1, ELMindex(ip,LEVELtot)
          VAL= D(i)*W(i,P)
          do j= 1, INL(i)
            VAL= VAL + AL(j,i)*W(IAL(j,i),P)
          enddo 
          do j= 1, INU(i)
            VAL= VAL + AU(j,i)*W(IAU(j,i),P)
          enddo 
          W(i,Q)= VAL
        enddo
      enddo
!$omp end parallel do
!C===

!C
!C +---------------------+
!C | ALPHA= RHO / {p}{q} |
!C +---------------------+
!C===
      C1= 0.d0
!$omp parallel do private(ip,i) reduction(+:C1)
      do ip= 1, PEsmpTOT
      do i = ELMindex(ip,0)+1, ELMindex(ip,LEVELtot)
        C1= C1 + W(i,P)*W(i,Q)
      enddo
      enddo
!$omp end parallel do

      ALPHA= RHO / C1
!C===


!C
!C +----------------------+
!C | {x}= {x} + ALPHA*{p} |
!C | {r}= {r} - ALPHA*{q} |
!C +----------------------+
!C===
!$omp parallel do private(ip,i)
      do ip= 1, PEsmpTOT
      do i = ELMindex(ip,0)+1, ELMindex(ip,LEVELtot)
        X(i)  = X(i)   + ALPHA * W(i,P)
        W(i,R)= W(i,R) - ALPHA * W(i,Q)
      enddo
      enddo
!$omp end parallel do

      DNRM2= 0.d0
!$omp parallel do private(ip,i) reduction(+:DNRM2)
      do ip= 1, PEsmpTOT
      do i = ELMindex(ip,0)+1, ELMindex(ip,LEVELtot)
        DNRM2= DNRM2 + W(i,R)**2
      enddo
      enddo
!$omp end parallel do
!C===

      ERR = dsqrt(DNRM2/BNRM2)
      if (mod(L,100).eq.1) then
        write (*,'(i5,2(1pe16.6))') L, ERR
      endif

        if (ERR .lt. EPS) then
          IER = 0
          goto 900
         else
          RHO1 = RHO
        endif

      enddo
      IER = 1

  900 continue

      write (*,'(i5,2(1pe16.6))') L, ERR
      ITR= L
!      EPS= ERR

      deallocate (W)

      return

      end subroutine  solve_ICCG_mc
      end module     solver_ICCG_mc



